require 'bitcoin'
require 'base_x'
require 'cashaddress'
require 'eth'
require 'base58'

class CoinAddressValidators
  TOKENS = %w[ret erct]
  require 'coin-address-validators/version'

  attr_reader :coin

  def initialize(coin)
    @coin = coin
  end

  def valid?(address)
    return false if address.nil?
    inner_validator.valid?(address)
  end

  def parse_cash_addr(address)
    inner_validator.parse_cash_addr(address)
  end

  def self.[](coin)
    @singletons ||= {}
    @singletons[coin] ||= new(coin)
  end

  private

  def inner_validator
    @inner_validator ||= inner_validator_class.new
  end

  def inner_validator_class
    prefix = coin == "usdt" ? "btc" : coin
    prefix = TOKENS.include?(coin) ? "eth" : prefix
    CoinAddressValidators.const_get("#{prefix.capitalize}Validator")
  end

  class BtcValidator
    include Bitcoin::Util
    def valid?(address)
      valid_address?(address)
    end
  end

  class LtcValidator
    include Bitcoin::Util

    def with_network(network)
      old_network = Bitcoin.network[:project]
      Bitcoin.network = network
      yield
    ensure
      Bitcoin.network = old_network
    end

    def valid?(address)
      with_network(:litecoin) { valid_address?(address) }
    end
  end

  class XrpValidator
    XRPBASE58 = BaseX.new("rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1tuvAxyz")

    def valid?(address)
      hex = XRPBASE58.decode(address).bth
      addr = hex[0...-8].rjust(42, "0")
      checksum = hex[-8..-1]
      Digest::SHA256.hexdigest(Digest::SHA256.digest(addr.htb))[0...8] == checksum
    rescue BaseX::InvalidNumeral => e
      false
    end
  end

  class BchValidator
    include Bitcoin::Util
    def parse_cash_addr(address)
      if address.start_with?("bitcoincash:")
        [Cashaddress.to_legacy(address), true]
      else
        [Cashaddress.to_legacy("bitcoincash:#{address}"), true]
      end
    rescue Cashaddress::Error, TypeError
      [nil, false]
    end

    def valid?(address)
      return true if valid_address?(address)
      legacy_address, ok = parse_cash_addr(address)
      ok
    end
  end

  class EthValidator
    def valid?(address)
      Eth::Utils.valid_address? address
    end
  end

  class TrxValidator
    def valid?(address)
      return false unless address.is_a?(String)
      return false unless address.size == 34
      return false unless address.start_with?("T")
      addr = BaseX::BitcoinBase58.decode(address)
      return false unless addr.size == 25
      return false unless addr.unpack1("H*").start_with?("41")

      checksum = addr[-4..-1]
      real_addr = addr[0..-5]

      # TODO checksum properly
      # hash0 = sha256(real_addr)
      # hash1 = sha256(hash0)
      # checksum1 = hash1[0..3]
      # return (checksum[0] == checksum1[0] && checksum[1] == checksum1[1] && checksum[2] ==
      #   checksum1[2] && checksum[3] == checksum1[3])

      return address == to_base58_address(real_addr.unpack1("H*"))
    end

    def to_base58_address(hex)
      return "" unless hex

      sha256ed = sha256(sha256(hex))
      bytes = hex_str_to_bytes(hex) + hex_str_to_bytes(sha256ed).first(4)
      Base58.encode(bytes_to_str(bytes).to_i(16), :bitcoin)
    end

    def hex_str_to_bytes(hex_str)
      hex_str.scan(/../).map(&:hex)
    end

    def bytes_to_str(bytes)
      bytes.pack("c*").unpack1("H*")
    end

    def sha256(hex_str)
      Digest::SHA2.new(256).hexdigest([hex_str].pack("H*"))
    end
  end
end
